@@ -5,7 +5,6 @@
  * For license terms, see the file COPYING in this directory.
  */
 
-/* We need this for HAVE_STDARG_H, etc */
 #include "config.h"
 
 struct addrinfo;
@@ -39,12 +38,8 @@ struct addrinfo;
 #  include "trio/trio.h"
 #endif
 
-/* We need this for strstr */
-#if !defined(HAVE_STRSTR) && !defined(strstr)
-char *strstr(const char *, const char *);
-#endif
-
 #include "fm_strl.h"
+#include "uid_db.h"
 
 /* constants designating the various supported protocols */
 #define		P_AUTO		1
@@ -119,7 +114,7 @@ char *strstr(const char *, const char *);
 #define		MSGBUFSIZE	8192
 
 #define		NAMELEN		64	/* max username length */
-#define		PASSWORDLEN	64	/* max password length */
+#define		PASSWORDLEN	256	/* max password length */
 #define		DIGESTLEN	33	/* length of MD5 digest */
 
 /* exit code values */
@@ -215,7 +210,7 @@ UID_SEEN=	1,		/**< id was seen, but not deleted */
 UID_DELETED=	2,		/**< this message has been marked deleted */
 UID_EXPUNGED=	3		/**< this message has been expunged */
 };
-/*@}*/
+/**/
 
 
 struct query;
@@ -389,8 +384,7 @@ struct query
     int smtp_socket;		/* socket descriptor for SMTP connection */
     unsigned int uid;		/* UID of user to deliver to */
     struct idlist *skipped;	/* messages skipped on the mail server */
-    struct idlist *oldsaved, *newsaved;
-    struct idlist **oldsavedend;
+    struct uid_db oldsaved, newsaved;
     char lastdigest[DIGESTLEN];	/* last MD5 hash seen on this connection */
     char *folder;		/* folder currently being polled */
 
@@ -402,11 +396,11 @@ struct query
     struct query *next;		/* next query control block in chain */
 };
 
-struct msgblk			/* message header parsed for open_sink() */
+struct msgblk			/** message header parsed for open_sink() */
 {
-    char   		*headers;	/* raw message headers */
-    struct idlist	*recipients;	/* addressees */
-    char		return_path[HOSTLEN + USERNAMELEN + 4]; 
+    char   		*headers;	/**< raw message headers */
+    struct idlist	*recipients;	/**< addressees */
+    char		return_path[HOSTLEN + USERNAMELEN + 4]; /**< envelope sender */
     int			msglen;
     int			reallen;
 };
@@ -464,6 +458,7 @@ extern flag versioninfo;	/* emit only version info */
 extern char *user;		/* name of invoking user */
 extern char *home;		/* home directory of invoking user */
 extern char *fmhome;		/* fetchmail home directory */
+extern int at_home;		/* normally 1, but 0 if FETCHMAILHOME overrides it */
 extern int pass;		/* number of re-polling pass */
 extern flag configdump;		/* dump control blocks as Python dictionary */
 extern const char *fetchmailhost; /* either "localhost" or an FQDN */
@@ -517,6 +512,16 @@ void resetidletimeout(void);
 int do_protocol(struct query *, const struct method *);
 
 /* transact.c: transaction support */
+/** \ingroup gen_recv_split
+ * Data structure to cache data between gen_recv_split() calls,
+ * must be initialized before use by calling gen_recv_split_init(). */
+struct RecvSplit
+{
+    char prefix[100];		/**< prefix to match/repeat when splitting lines */
+    int cached;			/**< flag to record if we have data cached in \a buf */
+    char buf[MSGBUFSIZE];	/**< buffer for cached data */
+};
+
 void init_transact(const struct method *);
 int readheaders(int sock,
 		       long fetchlen,
@@ -530,12 +535,16 @@ void gen_send(int sock, const char *, ... )
     __attribute__ ((format (printf, 2, 3)))
     ;
 int gen_recv(int sock, char *buf, int size);
+void gen_recv_split_init(const char *prefix, struct RecvSplit *rs);
+int gen_recv_split(int sock, char *buf, int size, struct RecvSplit *rs);
 int gen_transact(int sock, const char *, ... )
     __attribute__ ((format (printf, 2, 3)))
     ;
 #else
 void gen_send();
 int gen_recv();
+void gen_recv_split_init();
+int gen_recv_split();
 int gen_transact();
 #endif
 extern struct msgblk msgblk;
@@ -579,8 +588,11 @@ void release_sink(struct query *);
 int close_sink(struct query *, struct msgblk *, flag);
 int open_warning_by_mail(struct query *);
 #if defined(HAVE_STDARG_H)
-void stuff_warning(const char *, struct query *, const char *, ... )
-    __attribute__ ((format (printf, 3, 4)))
+void stuff_warning(const char *,
+                   struct query *,
+                   const char *pfx,
+                   const char *fmt, ...)
+    __attribute__ ((format (printf, 4, 5)))
     ;
 #else
 void stuff_warning();
@@ -622,8 +634,8 @@ int prc_parse_file(const char *, const flag);
 int prc_filecheck(const char *, const flag);
 
 /* base64.c */
-void to64frombits(char *, const void *, int);
-int from64tobits(void *, const char *, int maxlen);
+int to64frombits(char *, const void *, int inlen, size_t outlen);
+int from64tobits(void *, const char *, int mxoutlen);
 
 /* unmime.c */
 /* Bit-mask returned by MimeBodyType */
@@ -662,11 +674,12 @@ int do_otp(int sock, const char *command, struct query *ctl);
 extern char currentwd[1024], rcfiledir[1024];
 
 struct query *hostalloc(struct query *); 
-int parsecmdline (int, char **, struct runctl *, struct query *);
+int parsecmdline (int, char **, struct runctl *, struct query *, flag *);
 char *prependdir (const char *, const char *);
 char *MD5Digest (unsigned const char *);
 void hmac_md5 (const unsigned char *, size_t, const unsigned char *, size_t, unsigned char *, size_t);
 int POP3_auth_rpa(char *, char *, int socket);
+#define RETSIGTYPE void
 typedef RETSIGTYPE (*SIGHANDLERTYPE) (int);
 void deal_with_sigchld(void);
 RETSIGTYPE null_signal_handler(int sig);
@@ -757,9 +770,9 @@ int servport(const char *service);
 int fm_getaddrinfo(const char *node, const char *serv, const struct addrinfo *hints, struct addrinfo **res);
 void fm_freeaddrinfo(struct addrinfo *ai);
 
-/* prototypes from tls.c */
-int maybe_tls(struct query *ctl);
-int must_tls(struct query *ctl);
+/* prototypes from starttls.c */
+int maybe_starttls(struct query *ctl);
+int must_starttls(struct query *ctl);
 
 /* prototype from rfc822valid.c */
 int rfc822_valid_msgid(const unsigned char *);
