@@ -11,6 +11,8 @@
  */
 
 #include  "config.h"
+#include  "fetchmail.h"
+
 #include  <stdio.h>
 #include  <errno.h>
 #include  <string.h>
@@ -32,8 +34,6 @@
 #include  <ctype.h>
 #include  <langinfo.h>
 
-#include  "fetchmail.h"
-
 /* for W* macros after pclose() */
 #define _USE_BSD
 #include <sys/types.h>
@@ -76,7 +76,7 @@ static void smtp_rset(struct query *ctl)
     }
 }
 
-int smtp_open(struct query *ctl)
+int smtp_setup(struct query *ctl)
 /* try to open a socket to the appropriate SMTP server for this query */ 
 {
     /* maybe it's time to close the socket in order to force delivery */
@@ -114,7 +114,7 @@ int smtp_open(struct query *ctl)
 	 */
 	struct idlist	*idp;
 	const char *id_me = run.invisible ? ctl->server.truename : fetchmailhost;
-	int oldphase = phase;
+	int oldphase;
 	char *parsed_host = NULL;
 
 	errno = 0;
@@ -131,7 +131,7 @@ int smtp_open(struct query *ctl)
 	for (idp = ctl->smtphunt; idp; idp = idp->next)
 	{
 	    char	*cp;
-	    char	*portnum = SMTP_PORT;
+	    const char	*portnum = SMTP_PORT;
 
 	    ctl->smtphost = idp->id;  /* remember last host tried. */
 	    if (ctl->smtphost[0]=='/')
@@ -245,7 +245,8 @@ static void sanitize(char *s)
 char *rcpt_address(struct query *ctl, const char *id,
 			  int usesmtpname)
 {
-    static char addr[HOSTLEN+USERNAMELEN+1];
+    static char addr[HOSTLEN+USERNAMELEN+1000];
+
     if (strchr(id, '@'))
     {
 	snprintf(addr, sizeof (addr), "%s", id);
@@ -262,12 +263,13 @@ char *rcpt_address(struct query *ctl, const char *id,
 }
 
 static int send_bouncemail(struct query *ctl, struct msgblk *msg,
-			   int userclass, char *message /* should have \r\n at the end */,
+			   int userclass, const char *message /* should have \r\n at the end */,
 			   int nerrors, char *errors[])
 /* bounce back an error report a la RFC 1892 */
 {
     char daemon_name[15 + HOSTLEN] = "MAILER-DAEMON@";
-    char boundary[BUFSIZ], *bounce_to;
+    char boundary[BUFSIZ];
+    const char *bounce_to;
     int sock;
     static char *fqdn_of_host = NULL;
     const char *md1 = "MAILER-DAEMON", *md2 = "MAILER-DAEMON@";
@@ -440,23 +442,9 @@ static int handle_smtp_report(struct query *ctl, struct msgblk *msg)
 {
     int smtperr = atoi(smtp_response);
     char *responses[1];
-    struct idlist *walk;
-    int found = 0;
 
     responses[0] = xstrdup(smtp_response);
 
-#ifdef __UNUSED__
-    /*
-     * Don't do this!  It can really mess you up if, for example, you're
-     * reporting an error with a single RCPT TO address among several;
-     * RSET discards the message body and it doesn't get sent to the
-     * valid recipients.
-     */
-    smtp_rset(ctl);    /* stay on the safe side */
-    if (outlevel >= O_DEBUG)
-	report(stdout, GT_("Saved error is still %d\n"), smtperr);
-#endif /* __UNUSED */
-
     /*
      * Note: send_bouncemail message strings are not made subject
      * to gettext translation because (a) they're going to be 
@@ -465,15 +453,7 @@ static int handle_smtp_report(struct query *ctl, struct msgblk *msg)
      * messages, which are probably in English (none of the
      * MTAs I know about are internationalized).
      */
-    for( walk = ctl->antispam; walk; walk = walk->next )
-        if ( walk->val.status.num == smtperr ) 
-	{ 
-		found=1;
-		break;
-	}
-
-    /* if (str_find(&ctl->antispam, smtperr)) */
-    if ( found )
+    if (str_find(&ctl->antispam, smtperr))
     {
 	/*
 	 * SMTP listener explicitly refuses to deliver mail
@@ -541,12 +521,6 @@ static int handle_smtp_report(struct query *ctl, struct msgblk *msg)
 	 * (b) we wouldn't want spammers to get confirmation that
 	 * this address is live, anyway.
 	 */
-#ifdef __DONT_FEED_THE_SPAMMERS__
-	if (run.bouncemail)
-	    send_bouncemail(ctl, msg, XMIT_ACCEPT,
-			"Invalid address in MAIL FROM (SMTP error 553).\r\n", 
-			1, responses);
-#endif /* __DONT_FEED_THE_SPAMMERS__ */
 	free(responses[0]);
 	return(PS_REFUSED);
 
@@ -625,10 +599,6 @@ static int handle_smtp_report_without_bounce(struct query *ctl, struct msgblk *m
 	return(PS_REFUSED);
 
     case 553: /* invalid sending domain */
-#ifdef __DONT_FEED_THE_SPAMMERS__
-	if (run.bouncemail)
-	    return(PS_SUCCESS);
-#endif /* __DONT_FEED_THE_SPAMMERS__ */
 	return(PS_REFUSED);
 
     default:
@@ -648,6 +618,9 @@ int stuffline(struct query *ctl, char *buf)
     int	n, oldphase;
     char *last;
 
+    if (!buf)
+	return -1;
+
     /* The line may contain NUL characters. Find the last char to use
      * -- the real line termination is the sequence "\n\0".
      */
@@ -784,7 +757,7 @@ static int open_bsmtp_sink(struct query *ctl, struct msgblk *msg,
 
     if (fflush(sinkfp) || ferror(sinkfp))
     {
-	report(stderr, GT_("BSMTP preamble write failed.\n"));
+	report(stderr, GT_("BSMTP preamble write failed: %s.\n"), strerror(errno));
 	return(PS_BSMTP);
     }
 
@@ -850,7 +823,7 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
     const char	*ap;
     struct	idlist *idp;
     char		options[MSGBUFSIZE]; 
-    char		addr[HOSTLEN+USERNAMELEN+1];
+    char		addr[HOSTLEN+USERNAMELEN+1000];
 #ifdef EXPLICIT_BOUNCE_ON_BAD_ADDRESS
     char		**from_responses;
 #endif /* EXPLICIT_BOUNCE_ON_BAD_ADDRESS */
@@ -1129,9 +1102,9 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	 */
 	nameslen = 0;
 	for (idp = msg->recipients; idp; idp = idp->next)
-	    if ((idp->val.status.mark == XMIT_ACCEPT))
+	    if (idp->val.status.mark == XMIT_ACCEPT)
 		nameslen += (strlen(idp->id) + 1);	/* string + ' ' */
-	if ((*good_addresses == 0))
+	if (*good_addresses == 0)
 	    nameslen = strlen(run.postmaster);
 
 	names = (char *)xmalloc(nameslen + 1);	/* account for '\0' */
@@ -1190,18 +1163,28 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	for (dp = after, sp = before; (*dp = *sp); dp++, sp++) {
 	    if (sp[0] != '%')	continue;
 
+	    if (sp > before && sp[-1] == '\'') {
+		report(stderr, GT_("MDA option contains single-quoted %%%c expansion.\n"), sp[1]);
+		report(stderr, GT_("Refusing to deliver. Check the manual and fix your mda option.\n"));
+		free(before);
+		free(after);
+		if (from) free(from);
+		if (names) free(names);
+		return PS_SYNTAX;
+	    }
+
 	    /* need to expand? BTW, no here overflow, because in
 	    ** the worst case (end of string) sp[1] == '\0' */
 	    if (sp[1] == 's' || sp[1] == 'T') {
 		*dp++ = '\'';
-		strcpy(dp, names);
+		if (names) strcpy(dp, names);
 		dp += nameslen;
 		*dp++ = '\'';
 		sp++;	/* position sp over [sT] */
 		dp--;	/* adjust dp */
 	    } else if (sp[1] == 'F') {
 		*dp++ = '\'';
-		strcpy(dp, from);
+		if (from) strcpy(dp, from);
 		dp += fromlen;
 		*dp++ = '\'';
 		sp++;	/* position sp over F */
@@ -1235,7 +1218,10 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
      * under all BSDs and Linux)
      */
     orig_uid = getuid();
-    seteuid(ctl->uid);
+    if (seteuid(ctl->uid)) {
+	report(stderr, GT_("Cannot switch effective user id to %ld: %s\n"), (long)ctl->uid, strerror(errno));
+	return PS_IOERR;
+    }
 #endif /* HAVE_SETEUID */
 
     sinkfp = popen(before, "w");
@@ -1244,7 +1230,10 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 
 #ifdef HAVE_SETEUID
     /* this will fail quietly if we didn't start as root */
-    seteuid(orig_uid);
+    if (seteuid(orig_uid)) {
+	report(stderr, GT_("Cannot switch effective user id back to original %ld: %s\n"), (long)orig_uid, strerror(errno));
+	return PS_IOERR;
+    }
 #endif /* HAVE_SETEUID */
 
     if (!sinkfp)
@@ -1269,6 +1258,8 @@ int open_sink(struct query *ctl, struct msgblk *msg,
 {
     *bad_addresses = *good_addresses = 0;
 
+    if (want_progress() && outlevel >= O_VERBOSE && !ctl->mda && !ctl->bsmtp) puts("");
+
     if (ctl->bsmtp)		/* dump to a BSMTP batch file */
 	return(open_bsmtp_sink(ctl, msg, good_addresses, bad_addresses));
     /* 
@@ -1276,7 +1267,7 @@ int open_sink(struct query *ctl, struct msgblk *msg,
      * open a socket fails, fall through to attempt delivery via
      * local MDA.
      */
-    else if (!ctl->mda && smtp_open(ctl) != -1)
+    else if (!ctl->mda && smtp_setup(ctl) != -1)
 	return(open_smtp_sink(ctl, msg, good_addresses, bad_addresses));
 
     /*
@@ -1347,8 +1338,31 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 /* perform end-of-message actions on the current output sink */
 {
     int smtp_err;
-    if (ctl->mda)
-    {
+
+    if (want_progress() && outlevel >= O_VERBOSE && !ctl->mda && !ctl->bsmtp) puts("");
+
+    if (ctl->bsmtp && sinkfp) {
+	int error, oerrno;
+
+	/* implicit disk-full check here... */
+	fputs(".\r\n", sinkfp);
+	error = ferror(sinkfp);
+	oerrno = errno;
+	if (strcmp(ctl->bsmtp, "-"))
+	{
+	    if (fclose(sinkfp) == EOF) {
+		error = 1;
+		oerrno = errno;
+	    }
+	    sinkfp = (FILE *)NULL;
+	}
+	if (error)
+	{
+	    report(stderr, 
+		   GT_("Message termination or close of BSMTP file failed: %s\n"), strerror(oerrno));
+	    return(FALSE);
+	}
+    } else if (ctl->mda) {
 	int rc = 0, e = 0, e2 = 0, err = 0;
 
 	/* close the delivery pipe, we'll reopen before next message */
@@ -1364,8 +1378,6 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 	    e = errno;
 	    sinkfp = (FILE *)NULL;
 	}
-	else
-	    rc = e = 0;
 
 	deal_with_sigchld(); /* Restore SIGCHLD handling to reap zombies */
 
@@ -1388,25 +1400,6 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 	    return(FALSE);
 	}
     }
-    else if (ctl->bsmtp && sinkfp)
-    {
-	int error;
-
-	/* implicit disk-full check here... */
-	fputs(".\r\n", sinkfp);
-	error = ferror(sinkfp);
-	if (strcmp(ctl->bsmtp, "-"))
-	{
-	    if (fclose(sinkfp) == EOF) error = 1;
-	    sinkfp = (FILE *)NULL;
-	}
-	if (error)
-	{
-	    report(stderr, 
-		   GT_("Message termination or close of BSMTP file failed\n"));
-	    return(FALSE);
-	}
-    }
     else if (forward)
     {
 	/* write message terminator */
@@ -1486,8 +1479,16 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 		    }
 		    if (smtp_err != SM_OK)
 		    {
-			responses[errors] = xstrdup(smtp_response);
-			errors++;
+			/*
+			 * amavis returns the SMTP code from the recieving
+			 * host after the DATA-DOT. So we need to compare the
+			 * response to the antispam option here instead.
+			 */
+			if (handle_smtp_report(ctl, msg) != PS_REFUSED) {
+			    /* Only count an error if the message was not refused */
+			    responses[errors] = xstrdup(smtp_response);
+			    errors++;
+			}
 		    }
 		}
 
@@ -1558,12 +1559,12 @@ int open_warning_by_mail(struct query *ctl)
     else				/* send to postmaster  */
 	status = open_sink(ctl, &reply, &good, &bad);
     if (status == 0) {
-	stuff_warning(NULL, ctl, "From: FETCHMAIL-DAEMON@%s",
+	stuff_warning(NULL, ctl, "From: ", "FETCHMAIL-DAEMON@%s",
 		ctl->smtpaddress ? ctl->smtpaddress : fetchmailhost);
-	stuff_warning(NULL, ctl, "Date: %s", rfc822timestamp());
-	stuff_warning(NULL, ctl, "MIME-Version: 1.0");
-	stuff_warning(NULL, ctl, "Content-Transfer-Encoding: 8bit");
-	stuff_warning(NULL, ctl, "Content-Type: text/plain; charset=\"%s\"", iana_charset);
+	stuff_warning(NULL, ctl, "Date: ", "%s", rfc822timestamp());
+	stuff_warning(NULL, ctl, "MIME-Version: ", "1.0");
+	stuff_warning(NULL, ctl, "Content-Transfer-Encoding: ", "8bit");
+	stuff_warning(NULL, ctl, "Content-Type: ", "text/plain; charset=\"%s\"", iana_charset);
     }
     return(status);
 }
@@ -1573,19 +1574,22 @@ int open_warning_by_mail(struct query *ctl)
  * a header line) as per RFC-2047 using rfc2047charset as the character
  * set field */
 #if defined(HAVE_STDARG_H)
-void stuff_warning(const char *rfc2047charset, struct query *ctl, const char *fmt, ... )
+void stuff_warning(const char *rfc2047charset, struct query *ctl, const char *pfx, const char *fmt, ... )
 #else
-void stuff_warning(rfc2047charset, ctl, fmt, va_alist)
+void stuff_warning(rfc2047charset, ctl, pfx, fmt, va_alist)
 const char *charset;
+const char *pfx;	/* constant, non-translated prefix (such as "Subject: ") */
 struct query *ctl;
 const char *fmt;	/* printf-style format */
 va_dcl
 #endif
 {
     /* make huge -- i18n can bulk up error messages a lot */
-    char	buf[2*MSGBUFSIZE+4];
+    char	buf[3*MSGBUFSIZE+4];
     va_list ap;
 
+    snprintf(buf, sizeof(buf), "%s", pfx);
+
     /*
      * stuffline() requires its input to be writeable (for CR stripping),
      * so we needed to copy the message to a writeable buffer anyway in
@@ -1597,7 +1601,7 @@ va_dcl
 #else
     va_start(ap);
 #endif
-    vsnprintf(buf, sizeof(buf) - 2, fmt, ap);
+    vsnprintf(buf+strlen(buf), sizeof(buf)-strlen(buf)-2, fmt, ap);
     va_end(ap);
 
     snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "\r\n");
@@ -1613,7 +1617,7 @@ va_dcl
 void close_warning_by_mail(struct query *ctl, struct msgblk *msg)
 /* sign and send mailed warnings */
 {
-    stuff_warning(NULL, ctl, GT_("-- \nThe Fetchmail Daemon"));
+    stuff_warning(NULL, ctl, "-- \n", GT_("The Fetchmail Daemon"));
     close_sink(ctl, msg, TRUE);
 }
 
