@@ -106,7 +106,7 @@ static void SMTP_auth(int sock, char smtp_mode, char *username, char *password,
 		digest[9], digest[10], digest[11], digest[12], digest[13],
 		digest[14], digest[15]);
 
-		to64frombits(b64buf, tmp, strlen(tmp));
+		to64frombits(b64buf, tmp, strlen(tmp), sizeof b64buf);
 		SockPrintf(sock, "%s\r\n", b64buf);
 		SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
 	}
@@ -122,7 +122,7 @@ static void SMTP_auth(int sock, char smtp_mode, char *username, char *password,
 			if (tmp[c] == '^')
 				tmp[c] = '\0';
 		}
-		to64frombits(b64buf, tmp, len);
+		to64frombits(b64buf, tmp, len, sizeof b64buf);
 		SockPrintf(sock, "AUTH PLAIN %s\r\n", b64buf);
 		SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
 	}
@@ -144,7 +144,7 @@ static void SMTP_auth(int sock, char smtp_mode, char *username, char *password,
 			SMTP_auth_error(sock, GT_("Bad base64 reply from server.\n"));
 			return;
 		}
-		to64frombits(b64buf, username, strlen(username));
+		to64frombits(b64buf, username, strlen(username), sizeof b64buf);
 		SockPrintf(sock, "%s\r\n", b64buf);
 		SockRead(sock, smtp_response, sizeof(smtp_response) - 1);
 		strlcpy(tmp, smtp_response, sizeof(tmp));
@@ -159,7 +159,7 @@ static void SMTP_auth(int sock, char smtp_mode, char *username, char *password,
 			SMTP_auth_error(sock, GT_("Bad base64 reply from server.\n"));
 			return;
 		}
-		to64frombits(b64buf, password, strlen(password));
+		to64frombits(b64buf, password, strlen(password), sizeof b64buf);
 		SockPrintf(sock, "%s\r\n", b64buf);
 		SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
 	}
@@ -185,15 +185,16 @@ int SMTP_ehlo(int sock, char smtp_mode, const char *host, char *name, char *pass
   *opt = 0;
   while ((SockRead(sock, smtp_response, sizeof(smtp_response)-1)) != -1)
   {
-      int  n = strlen(smtp_response);
+      size_t n;
 
       set_timeout(0);
       (void)set_signal_handler(SIGALRM, alrmsave);
 
-      if (smtp_response[strlen(smtp_response)-1] == '\n')
-	  smtp_response[strlen(smtp_response)-1] = '\0';
-      if (smtp_response[strlen(smtp_response)-1] == '\r')
-	  smtp_response[strlen(smtp_response)-1] = '\0';
+      n = strlen(smtp_response);
+      if (n > 0 && smtp_response[n-1] == '\n')
+	  smtp_response[--n] = '\0';
+      if (n > 0 && smtp_response[n-1] == '\r')
+	  smtp_response[--n] = '\0';
       if (n < 4)
 	  return SM_ERROR;
       smtp_response[n] = '\0';
@@ -203,8 +204,7 @@ int SMTP_ehlo(int sock, char smtp_mode, const char *host, char *name, char *pass
 	  if (!strncasecmp(hp->name, smtp_response+4, strlen(hp->name))) {
 	      *opt |= hp->value;
 	      if (strncmp(hp->name, "AUTH ", 5) == 0)
-		strncpy(auth_response, smtp_response, sizeof(auth_response));
-		auth_response[sizeof(auth_response)-1] = '\0';
+		strlcpy(auth_response, smtp_response, sizeof(auth_response));
 	  }
       if ((smtp_response[0] == '1' || smtp_response[0] == '2' || smtp_response[0] == '3') && smtp_response[3] == ' ') {
 	  if (*opt & ESMTP_AUTH)
@@ -329,10 +329,9 @@ int SMTP_ok(int sock, char smtp_mode, int mintimeout)
 
 	n = strlen(reply);
 	if (n > 0 && reply[n-1] == '\n')
-	    n--;
+	    reply[--n] = '\0';
 	if (n > 0 && reply[n-1] == '\r')
-	    n--;
-	reply[n] = '\0';
+	    reply[--n] = '\0';
 
 	/* stomp over control characters */
 	for (i = reply; *i; i++)
