@@ -1,11 +1,16 @@
 /*
  * socket.c -- socket library functions
  *
- * Copyright 1998 by Eric S. Raymond.
+ * Copyright 1998 - 2004 by Eric S. Raymond.
+ * Copyright 2004 - 2019 by Matthias Andree.
+ * Contributions by Alexander Bluhm, Earl Chew, John Beck.
+
  * For license terms, see the file COPYING in this directory.
  */
 
 #include "config.h"
+#include "fetchmail.h"
+
 #include <stdio.h>
 #include <errno.h>
 #include <string.h>
@@ -15,11 +20,7 @@
 #endif /* HAVE_MEMORY_H */
 #include <sys/types.h>
 #include <sys/stat.h>
-#ifndef HAVE_NET_SOCKET_H
 #include <sys/socket.h>
-#else
-#include <net/socket.h>
-#endif
 #include <sys/un.h>
 #include <netinet/in.h>
 #ifdef HAVE_ARPA_INET_H
@@ -49,10 +50,10 @@
 #endif
 
 #include "socket.h"
-#include "fetchmail.h"
 #include "getaddrinfo.h"
 #include "i18n.h"
 #include "sdump.h"
+#include "uid_db.h"
 
 /* Defines to allow BeOS and Cygwin to play nice... */
 #ifdef __BEOS__
@@ -111,7 +112,7 @@ static char *const *parse_plugin(const char *plugin, const char *host, const cha
 		return NULL;
 	}
 
-	while (plugin_copy_offset < plugin_copy_len)
+	while (plugin_offset < plugin_len && plugin_copy_offset < plugin_copy_len)
 	{	if ((plugin[plugin_offset] == '%') && (plugin[plugin_offset + 1] == 'h'))
 		{	strcpy(plugin_copy + plugin_copy_offset, host);
 			plugin_offset += 2;
@@ -130,6 +131,8 @@ static char *const *parse_plugin(const char *plugin, const char *host, const cha
 	}
 	plugin_copy[plugin_copy_len] = 0;
 
+	/* XXX FIXME - is this perhaps a bit too simplistic to chop down the argument strings without any respect to quoting?
+	 * better write a generic function that tracks arguments instead... */
 	argvec = (char **)malloc(s);
 	if (!argvec)
 	{
@@ -204,7 +207,7 @@ static int handle_plugin(const char *host,
 #endif /* HAVE_SOCKETPAIR */
 
 /** Set socket to SO_KEEPALIVE. \return 0 for success. */
-int SockKeepalive(int sock) {
+static int SockKeepalive(int sock) {
     int keepalive = 1;
     return setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &keepalive, sizeof keepalive);
 }
@@ -215,7 +218,7 @@ int UnixOpen(const char *path)
     struct sockaddr_un ad;
     memset(&ad, 0, sizeof(ad));
     ad.sun_family = AF_UNIX;
-    strncpy(ad.sun_path, path, sizeof(ad.sun_path)-1);
+    strlcpy(ad.sun_path, path, sizeof(ad.sun_path));
 
     sock = socket( AF_UNIX, SOCK_STREAM, 0 );
     if (sock < 0)
@@ -224,7 +227,7 @@ int UnixOpen(const char *path)
 	return -1;
     }
 
-    /* Socket opened saved. Usefull if connect timeout 
+    /* Socket opened saved. Useful if connect timeout
      * because it can be closed.
      */
     mailserver_socket_temp = sock;
@@ -245,7 +248,7 @@ int UnixOpen(const char *path)
 }
 
 int SockOpen(const char *host, const char *service,
-	     const char *plugin, struct addrinfo **ai0)
+	     const char *plugin, struct addrinfo **ai_in)
 {
     struct addrinfo *ai, req;
     int i, acterr = 0;
@@ -263,7 +266,7 @@ int SockOpen(const char *host, const char *service,
     req.ai_flags = AI_ADDRCONFIG;
 #endif
 
-    i = fm_getaddrinfo(host, service, &req, ai0);
+    i = fm_getaddrinfo(host, service, &req, ai_in);
     if (i) {
 	report(stderr, GT_("getaddrinfo(\"%s\",\"%s\") error: %s\n"),
 		host, service, gai_strerror(i));
@@ -275,7 +278,7 @@ int SockOpen(const char *host, const char *service,
     /* NOTE a Linux bug here - getaddrinfo will happily return 127.0.0.1
      * twice if no IPv6 is configured */
     i = -1;
-    for (ord = 0, ai = *ai0; ai; ord++, ai = ai->ai_next) {
+    for (ord = 0, ai = *ai_in; ai; ord++, ai = ai->ai_next) {
 	char buf[256]; /* hostname */
 	char pb[256];  /* service name */
 	int gnie;      /* getnameinfo result code */
@@ -335,8 +338,8 @@ int SockOpen(const char *host, const char *service,
 	break;
     }
 
-    fm_freeaddrinfo(*ai0);
-    *ai0 = NULL;
+    fm_freeaddrinfo(*ai_in);
+    *ai_in = NULL;
 
     if (i == -1) {
 	report(stderr, GT_("Connection errors for this poll:\n%s"), errbuf);
@@ -372,12 +375,33 @@ va_dcl {
 }
 
 #ifdef SSL_ENABLE
+/* OPENSSL_NO_SSL_INTERN: 
+   transitional feature for OpenSSL 1.0.1 up to and excluding 1.1.0 
+   to make sure we do not access internal structures! */
+#define OPENSSL_NO_SSL_INTERN 1
+#define OPENSSL_NO_DEPRECATED 23
 #include <openssl/ssl.h>
 #include <openssl/err.h>
 #include <openssl/pem.h>
 #include <openssl/x509v3.h>
 #include <openssl/rand.h>
 
+#define fm_MIN_OPENSSL_VER 0x1000200fL
+
+#ifdef LIBRESSL_VERSION_NUMBER
+#pragma message "WARNING - LibreSSL is unsupported. Use at your own risk."
+#endif
+
+#if OPENSSL_VERSION_NUMBER < fm_MIN_OPENSSL_VER
+#error Your OpenSSL version must be at least 1.0.2 release. Older OpenSSL versions are unsupported.
+#else
+/*
+#define __fm_ossl_ver(x) #x
+#define _fm_ossl_ver(x) __fm_ossl_ver(x)
+#pragma message "Building with OpenSSL headers version " _fm_ossl_ver(OPENSSL_VERSION_NUMBER) ", " OPENSSL_VERSION_TEXT
+*/
+#endif
+
 static void report_SSL_errors(FILE *stream)
 {
     unsigned long err;
@@ -597,7 +621,7 @@ SSL *SSLGetContext( int sock )
 /* ok_return (preverify_ok) is 1 if this stage of certificate verification
    passed, or 0 if it failed. This callback lets us display informative
    errors, and perform additional validation (e.g. CN matches) */
-static int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
+static int SSL_verify_callback(int ok_return, X509_STORE_CTX *ctx, int strict)
 {
 #define SSLverbose (((outlevel) >= O_DEBUG) || ((outlevel) >= O_VERBOSE && (depth) == 0)) 
 	char buf[257];
@@ -617,6 +641,12 @@ static int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
 	subj = X509_get_subject_name(x509_cert);
 	issuer = X509_get_issuer_name(x509_cert);
 
+	if (outlevel >= O_DEBUG) {
+		if (SSLverbose)
+			report(stdout, GT_("SSL verify callback depth %d: preverify_ok == %d, err = %d, %s\n"),
+					depth, ok_return, err, X509_verify_cert_error_string(err));
+	}
+
 	if (outlevel >= O_VERBOSE) {
 		if (depth == 0 && SSLverbose)
 			report(stdout, GT_("Server certificate:\n"));
@@ -822,7 +852,7 @@ static int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
 	_verify_ok &= ok_return;
 	if (!strict)
 		ok_return = 1;
-	return (ok_return);
+	return ok_return;
 }
 
 static int SSL_nock_verify_callback( int ok_return, X509_STORE_CTX *ctx )
@@ -866,6 +896,157 @@ static const char *SSLCertGetCN(const char *mycert,
 	return ret;
 }
 
+#if defined(LIBRESSL_VERSION_NUMBER) || OPENSSL_VERSION_NUMBER < 0x1010000fL
+/* OSSL_proto_version_logic for OpenSSL 1.0.x and LibreSSL */
+static int OSSL10X_proto_version_logic(int sock, const char **myproto, int *avoid_ssl_versions)
+{
+	if (!*myproto) {
+	    *myproto = "auto";
+	}
+
+	if (!strcasecmp("ssl3", *myproto)) {
+#if (HAVE_DECL_SSLV3_CLIENT_METHOD > 0) && (0 == OPENSSL_NO_SSL3 + 0)
+		_ctx[sock] = SSL_CTX_new(SSLv3_client_method());
+		*avoid_ssl_versions &= ~SSL_OP_NO_SSLv3;
+#else
+		report(stderr, GT_("Your OpenSSL version does not support SSLv3.\n"));
+		return -1;
+#endif
+	} else if (!strcasecmp("ssl3+", *myproto)) {
+		*avoid_ssl_versions &= ~SSL_OP_NO_SSLv3;
+		*myproto = NULL;
+	} else if (!strcasecmp("tls1", *myproto)) {
+		_ctx[sock] = SSL_CTX_new(TLSv1_client_method());
+	} else if (!strcasecmp("tls1+", *myproto)) {
+		*myproto = NULL;
+#if defined(TLS1_1_VERSION)
+	} else if (!strcasecmp("tls1.1", *myproto)) {
+		_ctx[sock] = SSL_CTX_new(TLSv1_1_client_method());
+	} else if (!strcasecmp("tls1.1+", *myproto)) {
+		*myproto = NULL;
+		*avoid_ssl_versions |= SSL_OP_NO_TLSv1;
+#else
+	} else if(!strcasecmp("tls1.1",*myproto) || !strcasecmp("tls1.1+", *myproto)) {
+		report(stderr, GT_("Your OpenSSL version does not support TLS v1.1.\n"));
+		return -1;
+#endif
+#if defined(TLS1_2_VERSION)
+	} else if (!strcasecmp("tls1.2", *myproto)) {
+		_ctx[sock] = SSL_CTX_new(TLSv1_2_client_method());
+	} else if (!strcasecmp("tls1.2+", *myproto)) {
+		*myproto = NULL;
+		*avoid_ssl_versions |= SSL_OP_NO_TLSv1 | SSL_OP_NO_TLSv1_1;
+#else
+	} else if(!strcasecmp("tls1.2",*myproto) || !strcasecmp("tls1.2+", *myproto)) {
+		report(stderr, GT_("Your OpenSSL version does not support TLS v1.2.\n"));
+		return -1;
+#endif
+#if defined(TLS1_3_VERSION)
+	} else if (!strcasecmp("tls1.3", *myproto)) {
+		_ctx[sock] = SSL_CTX_new(TLSv1_3_client_method());
+	} else if (!strcasecmp("tls1.3+", *myproto)) {
+		*myproto = NULL;
+		*avoid_ssl_versions |= SSL_OP_NO_TLSv1 | SSL_OP_NO_TLSv1_1 | SSL_OP_NO_TLSv1_2;
+#else
+	} else if(!strcasecmp("tls1.3",*myproto) || !strcasecmp("tls1.3+", *myproto)) {
+		report(stderr, GT_("Your OpenSSL version does not support TLS v1.3.\n"));
+		return -1;
+#endif
+	} else if (!strcasecmp("ssl23", *myproto)
+	        || 0 == strcasecmp("auto", *myproto))
+	{
+		*myproto = NULL;
+	} else {
+		report(stderr,
+		        GT_("Invalid SSL protocol '%s' specified, using default autoselect (auto).\n"),
+		        *myproto);
+		*myproto = NULL;
+	}
+	return 0;
+}
+#define OSSL_proto_version_logic(a,b,c) OSSL10X_proto_version_logic((a),(b),(c))
+#undef OSSL110_API
+#else
+/* implementation for OpenSSL 1.1.0 */
+#define OSSL110_API 1
+static int OSSL110_proto_version_logic(int sock, const char **myproto,
+        int *avoid_ssl_versions)
+{
+	/* NOTE - this code MUST NOT set myproto to NULL, else the
+	 * SSL_...set_..._proto_version() call becomes ineffective. */
+	_ctx[sock] = SSL_CTX_new(TLS_client_method());
+	SSL_CTX_set_min_proto_version(_ctx[sock], TLS1_VERSION);
+
+	if (!*myproto) {
+	    *myproto = "auto";
+	}
+
+	if (!strcasecmp("ssl3", *myproto)) {
+#if (0 == OPENSSL_NO_SSL3 + 0)
+		SSL_CTX_set_min_proto_version(_ctx[sock], SSL3_VERSION);
+		SSL_CTX_set_max_proto_version(_ctx[sock], SSL3_VERSION);
+		*avoid_ssl_versions &= ~SSL_OP_NO_SSLv3;
+#else
+		report(stderr, GT_("Your OpenSSL version does not support SSLv3.\n"));
+		return -1;
+#endif
+	} else if (!strcasecmp("ssl3+", *myproto)) {
+		SSL_CTX_set_min_proto_version(_ctx[sock], SSL3_VERSION);
+		*avoid_ssl_versions &= ~SSL_OP_NO_SSLv3;
+	} else if (!strcasecmp("tls1", *myproto)) {
+		SSL_CTX_set_max_proto_version(_ctx[sock], TLS1_VERSION);
+	} else if (!strcasecmp("tls1+", *myproto)) {
+		/* do nothing, min_proto_version is already at TLS1_VERSION */
+#if defined(TLS1_1_VERSION)
+	} else if (!strcasecmp("tls1.1", *myproto)) {
+		SSL_CTX_set_min_proto_version(_ctx[sock], TLS1_1_VERSION);
+		SSL_CTX_set_max_proto_version(_ctx[sock], TLS1_1_VERSION);
+	} else if (!strcasecmp("tls1.1+", *myproto)) {
+		SSL_CTX_set_min_proto_version(_ctx[sock], TLS1_1_VERSION);
+#else
+	} else if(!strcasecmp("tls1.1",*myproto) || !strcasecmp("tls1.1+", *myproto)) {
+		report(stderr, GT_("Your OpenSSL version does not support TLS v1.1.\n"));
+		return -1;
+#endif
+#if defined(TLS1_2_VERSION)
+	} else if (!strcasecmp("tls1.2", *myproto)) {
+		SSL_CTX_set_min_proto_version(_ctx[sock], TLS1_2_VERSION);
+		SSL_CTX_set_max_proto_version(_ctx[sock], TLS1_2_VERSION);
+	} else if (!strcasecmp("tls1.2+", *myproto)) {
+		SSL_CTX_set_min_proto_version(_ctx[sock], TLS1_2_VERSION);
+#else
+	} else if(!strcasecmp("tls1.2",*myproto) || !strcasecmp("tls1.2+", *myproto)) {
+		report(stderr, GT_("Your OpenSSL version does not support TLS v1.2.\n"));
+		return -1;
+#endif
+#if defined(TLS1_3_VERSION)
+	} else if (!strcasecmp("tls1.3", *myproto)) {
+		SSL_CTX_set_min_proto_version(_ctx[sock], TLS1_3_VERSION);
+		SSL_CTX_set_max_proto_version(_ctx[sock], TLS1_3_VERSION);
+	} else if (!strcasecmp("tls1.3+", *myproto)) {
+		SSL_CTX_set_min_proto_version(_ctx[sock], TLS1_3_VERSION);
+#else
+	} else if(!strcasecmp("tls1.3",*myproto) || !strcasecmp("tls1.3+", *myproto)) {
+		report(stderr, GT_("Your OpenSSL version does not support TLS v1.3.\n"));
+		return -1;
+#endif
+	} else if (!strcasecmp("ssl23", *myproto)
+	        || 0 == strcasecmp("auto", *myproto))
+	{
+		/* do nothing */
+	} else {
+		/* This should not happen. */
+		report(stderr,
+		        GT_("Invalid SSL protocol '%s' specified, using default autoselect (auto).\n"),
+		        *myproto);
+		report(stderr, "fetchmail internal error in OSSL110_proto_version_logic\n");
+		abort();
+	}
+	return 0;
+}
+#define OSSL_proto_version_logic(a,b,c) OSSL110_proto_version_logic((a),(b),(c))
+#endif
+
 /* performs initial SSL handshake over the connected socket
  * uses SSL *ssl global variable, which is currently defined
  * in this file
@@ -876,11 +1057,28 @@ int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck
 {
         struct stat randstat;
         int i;
-	long sslopts = SSL_OP_ALL;
+	int avoid_ssl_versions = SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3;
+	long sslopts = SSL_OP_ALL | SSL_OP_SINGLE_DH_USE;
+	int ssle_connect = 0;
+	long ver;
 
+#ifndef OSSL110_API
 	SSL_load_error_strings();
 	SSL_library_init();
 	OpenSSL_add_all_algorithms(); /* see Debian Bug#576430 and manpage */
+	ver = SSLeay();
+#else
+	ver = OpenSSL_version_num();
+#endif
+
+	if (ver < OPENSSL_VERSION_NUMBER) {
+	    report(stderr, GT_("Loaded OpenSSL library %#lx older than headers %#lx, refusing to work.\n"), (long)ver, (long)(OPENSSL_VERSION_NUMBER));
+	    return -1;
+	}
+
+	if (ver > OPENSSL_VERSION_NUMBER && outlevel >= O_VERBOSE) {
+	    report(stdout, GT_("Loaded OpenSSL library %#lx newer than headers %#lx, trying to continue.\n"), (long)ver, (long)(OPENSSL_VERSION_NUMBER));
+	}
 
         if (stat("/dev/random", &randstat)  &&
             stat("/dev/urandom", &randstat)) {
@@ -905,30 +1103,23 @@ int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck
 
 	/* Make sure a connection referring to an older context is not left */
 	_ssl_context[sock] = NULL;
-	if(myproto) {
-		if(!strcasecmp("ssl2",myproto)) {
-#if HAVE_DECL_SSLV2_CLIENT_METHOD + 0 > 0
-			_ctx[sock] = SSL_CTX_new(SSLv2_client_method());
-#else
-			report(stderr, GT_("Your operating system does not support SSLv2.\n"));
-			return -1;
-#endif
-		} else if(!strcasecmp("ssl3",myproto)) {
-			_ctx[sock] = SSL_CTX_new(SSLv3_client_method());
-		} else if(!strcasecmp("tls1",myproto)) {
-			_ctx[sock] = SSL_CTX_new(TLSv1_client_method());
-		} else if (!strcasecmp("ssl23",myproto)) {
-			myproto = NULL;
-		} else {
-			report(stderr,GT_("Invalid SSL protocol '%s' specified, using default (SSLv23).\n"), myproto);
-			myproto = NULL;
-		}
+	{
+		int rc = OSSL_proto_version_logic(sock, &myproto, &avoid_ssl_versions);
+		if (rc) return rc;
 	}
-	if(!myproto) {
+	/* do not combine into an else { } as myproto may be nulled above! */
+	if (!myproto) {
+		/* SSLv23 is a misnomer and will in fact use the best
+		 available protocol, subject to SSL_OP_NO* constraints. */
 		_ctx[sock] = SSL_CTX_new(SSLv23_client_method());
 	}
+
 	if(_ctx[sock] == NULL) {
+		unsigned long ec = ERR_peek_last_error();
 		ERR_print_errors_fp(stderr);
+		if (ERR_GET_REASON(ec) == SSL_R_NULL_SSL_METHOD_PASSED) {
+		    report(stderr, GT_("Note that some distributions disable older protocol versions in weird non-standard ways. Try a newer protocol version.\n"));
+		}
 		return(-1);
 	}
 
@@ -938,7 +1129,7 @@ int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck
 		sslopts &= ~ SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS;
 	}
 
-	SSL_CTX_set_options(_ctx[sock], sslopts);
+	SSL_CTX_set_options(_ctx[sock], sslopts | avoid_ssl_versions);
 
 	if (certck) {
 		SSL_CTX_set_verify(_ctx[sock], SSL_VERIFY_PEER, SSL_ck_verify_callback);
@@ -986,6 +1177,36 @@ int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck
 	_verify_ok = 1;
 	_prev_err = -1;
 
+	/*
+	 * Support SNI, some servers (googlemail) appear to require it.
+	 */
+	{
+	    long r;
+	    r = SSL_set_tlsext_host_name(_ssl_context[sock], servercname);
+
+	    if (0 == r) {
+		/* handle error */
+		report(stderr, GT_("Warning: SSL_set_tlsext_host_name(%p, \"%s\") failed (code %#lx), trying to continue.\n"), (void *)_ssl_context[sock], servercname, r);
+		ERR_print_errors_fp(stderr);
+	    }
+	}
+
+	/* OpenSSL >= 1.0.2: set host name for verification */
+	/* XXX FIXME: do we need to change the function's signature and pass the akalist to
+	 * permit the other hostnames through SSL? */
+	/* https://wiki.openssl.org/index.php/Hostname_validation */
+	{
+	    int r;
+	    X509_VERIFY_PARAM *param = SSL_get0_param(_ssl_context[sock]);
+
+	    X509_VERIFY_PARAM_set_hostflags(param, X509_CHECK_FLAG_NO_PARTIAL_WILDCARDS);
+	    if (0 == (r = X509_VERIFY_PARAM_set1_host(param, servercname, strlen(servercname)))) {
+		report(stderr, GT_("Warning: X509_VERIFY_PARAM_set1_host(%p, \"%s\") failed (code %#x), trying to continue.\n"),
+			(void *)_ssl_context[sock], servercname, r);
+		ERR_print_errors_fp(stderr);
+	    }
+	}
+
 	if( mycert || mykey ) {
 
 	/* Ok...  He has a certificate file defined, so lets declare it.  If
@@ -1003,13 +1224,26 @@ int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck
 			free(*remotename);
 			*remotename = xstrdup(buffer);
 		}
-        	SSL_use_certificate_file(_ssl_context[sock], mycert, SSL_FILETYPE_PEM);
-        	SSL_use_RSAPrivateKey_file(_ssl_context[sock], mykey, SSL_FILETYPE_PEM);
+		SSL_use_certificate_file(_ssl_context[sock], mycert, SSL_FILETYPE_PEM);
+		SSL_use_RSAPrivateKey_file(_ssl_context[sock], mykey, SSL_FILETYPE_PEM);
 	}
 
 	if (SSL_set_fd(_ssl_context[sock], sock) == 0 
-	    || SSL_connect(_ssl_context[sock]) < 1) {
+	    || (ssle_connect = SSL_connect(_ssl_context[sock])) < 1) {
+		int e = errno;
+		unsigned long ssle_err_from_get_error = SSL_get_error(_ssl_context[sock], ssle_connect);
+		unsigned long ssle_err_from_queue = ERR_peek_error();
 		ERR_print_errors_fp(stderr);
+		if (SSL_ERROR_SYSCALL == ssle_err_from_get_error && 0 == ssle_err_from_queue) {
+		    if (0 == ssle_connect) {
+			/* FIXME: the next line was hacked in 6.4.0-rc1 so the translation strings don't change.
+			 * The %s could be merged to the inside of GT_(). */
+			report(stderr, "%s: %s", servercname, GT_("Server shut down connection prematurely during SSL_connect().\n"));
+		    } else if (ssle_connect < 0) {
+			report(stderr, "%s: ", servercname);
+			report(stderr, GT_("System error during SSL_connect(): %s\n"), e ? strerror(e) : GT_("handshake failed at protocol or connection level."));
+		    }
+		}
 		SSL_free( _ssl_context[sock] );
 		_ssl_context[sock] = NULL;
 		SSL_CTX_free(_ctx[sock]);
@@ -1017,6 +1251,24 @@ int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck
 		return(-1);
 	}
 
+	if (outlevel >= O_VERBOSE) {
+	    SSL_CIPHER const *sc;
+	    int bitsmax, bitsused;
+
+	    const char *vers;
+
+	    vers = SSL_get_version(_ssl_context[sock]);
+
+	    sc = SSL_get_current_cipher(_ssl_context[sock]);
+	    if (!sc) {
+		report (stderr, GT_("Cannot obtain current SSL/TLS cipher - no session established?\n"));
+	    } else {
+		bitsused = SSL_CIPHER_get_bits(sc, &bitsmax);
+		report(stdout, GT_("SSL/TLS: using protocol %s, cipher %s, %d/%d secret/processed bits\n"),
+			vers, SSL_CIPHER_get_name(sc), bitsused, bitsmax);
+	    }
+	}
+
 	/* Paranoia: was the callback not called as we expected? */
 	if (!_depth0ck) {
 		report(stderr, GT_("Certificate/fingerprint verification was somehow skipped!\n"));
